/****************************************************************************

  This file is part of the VRender library.
  Copyright (C) 2005 Cyril Soler (Cyril.Soler@imag.fr)
  Version 1.0.0, released on June 27, 2005.

  http://artis.imag.fr/Members/Cyril.Soler/VRender

  VRender is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation; either version 2 of the License, or (at your option) any later
  version.
  
  VRender is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with
  VRender; if not, write to the Free Software Foundation, Inc., 59 Temple
  Place, Suite 330, Boston, MA  02111-1307  USA

*****************************************************************************/

#ifndef _VRENDER_VECTOR3_H
#define _VRENDER_VECTOR3_H

#include <stdexcept>
#include <iostream>

namespace vrender
{
	template<class T> class T_NVector3 ;

	template<class T> class T_Vector3
	{
		public:
			// ---------------------------------------------------------------------------
			//! @name Constant
			//@{
			static const T_Vector3 null;
			static const T_Vector3 inf;
			//@}

			// ---------------------------------------------------------------------------
			//! @name Constructor(s) and destructor
			//@{
			T_Vector3 ();
			~T_Vector3 ();
			T_Vector3 (const T_Vector3<T>&);
			T_Vector3 (const T_NVector3<T>&);
			T_Vector3 (T,T,T);
			// I don't like this at all, because it allows things like 
			// T_NVector3(Point3), which we certainly don't want although the compiler
			// won't complain at all about it. This produces bugs very hard to find
			// (I got one by doing 
			//       T_NVector3 toto(ray.getOrigin()) 
			// instead of
			//       Point3 toto(ray.getOrigin()) 
			//
			template <class U> inline explicit T_Vector3 (const U&);
			//@}

			// ---------------------------------------------------------------------------
			//! @name Access methods
			//@{
			inline T  x() const { return _xyz[0]; }
			inline T  y() const { return _xyz[1]; }
			inline T  z() const { return _xyz[2]; }
			inline void  setX(T r) { _xyz[0] = r; }
			inline void  setY(T r) { _xyz[1] = r; }
			inline void  setZ(T r) { _xyz[2] = r; }
			inline void  setXYZ (T x,T y,T z) { _xyz[0] = x; _xyz[1] = y; _xyz[2] = z; }
			//@}

			// ---------------------------------------------------------------------------
			//! @name Assignment
			//@{
			inline T_Vector3& operator= (const T_Vector3<T>& u)  { _xyz[0] = u._xyz[0]; _xyz[1] = u._xyz[1]; _xyz[2] = u._xyz[2]; return *this; }
			T_Vector3& operator= (const T_NVector3<T>& u);
			//@}

#ifdef A_VIRER
			// ---------------------------------------------------------------------------
			//! @name Comparisons
			//@{
			template<class U> friend bool operator== (const T_Vector3<U>&,const T_Vector3<U>&);
			template<class U> friend bool operator!= (const T_Vector3<U>&,const T_Vector3<U>&);
			bool isEqualTo (const T_Vector3<T>&,T epsilon=Real_epsilon) const;
			bool isNull (T epsilon=Real_epsilon) const;
#endif
			//@}

			// ---------------------------------------------------------------------------
			//! @name Algebraic operations
			//@{
			inline T_Vector3<T>& operator+= (const T_Vector3<T>& v)
			{ 
				_xyz[0] += v._xyz[0]; 
				_xyz[1] += v._xyz[1]; 
				_xyz[2] += v._xyz[2]; 
				return *this;
			}

			inline T_Vector3<T>& operator-= (const T_Vector3<T>& v)
			{ 
				_xyz[0] -= v._xyz[0]; 
				_xyz[1] -= v._xyz[1]; 
				_xyz[2] -= v._xyz[2]; 
				return *this;
			}

			inline T_Vector3<T>& operator*= (T f) { _xyz[0] *= f; _xyz[1] *= f; _xyz[2] *= f; return *this;}
			inline T_Vector3<T>& operator/= (T f) { _xyz[0] /= f; _xyz[1] /= f; _xyz[2] /= f; return *this;}

			template<class U> friend T_Vector3<U> min(const T_Vector3<U>&,const T_Vector3<U>&) ;
			template<class U> friend T_Vector3<U> max(const T_Vector3<U>&,const T_Vector3<U>&) ;

			T_Vector3<T>& operator-= (const T_NVector3<T>&);
			T_Vector3<T>& operator+= (const T_NVector3<T>&);

			template<class U> friend T_Vector3<U> operator- (const T_Vector3<U>&);

			inline T_Vector3<T> operator+(const T_Vector3<T>& u) const 
			{ 
				return T_Vector3<T>(_xyz[0]+u._xyz[0],_xyz[1]+u._xyz[1],_xyz[2]+u._xyz[2]); 
			}
			inline T_Vector3<T> operator-(const T_Vector3<T>& u) const 
			{ 
				return T_Vector3<T>(_xyz[0]-u._xyz[0],_xyz[1]-u._xyz[1],_xyz[2]-u._xyz[2]); 
			}

			inline T    operator*(const T_Vector3<T>& u) const 
			{ 
				return _xyz[0]*u._xyz[0] + _xyz[1]*u._xyz[1] + _xyz[2]*u._xyz[2]; 
			}

			inline T_Vector3<T> operator^(const T_Vector3<T>& v) const 
			{ 
				return T_Vector3<T>(	_xyz[1]*v._xyz[2] - _xyz[2]*v._xyz[1],
											_xyz[2]*v._xyz[0] - _xyz[0]*v._xyz[2],
											_xyz[0]*v._xyz[1] - _xyz[1]*v._xyz[0]); 
			}

			T_Vector3<T> operator/ (T v) { return T_Vector3<T>(_xyz[0]/v,_xyz[1]/v,_xyz[2]/v); }
			T_Vector3<T> operator* (T v) { return T_Vector3<T>(_xyz[0]*v,_xyz[1]*v,_xyz[2]*v); }

			template<class U> friend T_Vector3<U> operator* (U,const T_Vector3<U>&);
			//@}

			// ---------------------------------------------------------------------------
			//! @name Metrics
			//@{
			T norm       () const;
			T squareNorm () const;
			T infNorm    () const; /// Should be used for most comparisons, for efficiency reasons.
			//@}
			// ---------------------------------------------------------------------------
			//! @name Stream overrides
			//@{
			template<class U> friend std::ostream& operator<< (std::ostream&,const T_Vector3<U>&);
			//@}

			T  operator[] (int i) const 
			{ 
				if((i < 0)||(i > 2))
					throw std::runtime_error("Out of bounds in T_Vector3::operator[]") ;

				return _xyz[i]; 
			}

			T& operator[] (int i) 
			{ 
				if((i < 0)||(i > 2))
					throw std::runtime_error("Out of bounds in T_Vector3::operator[]") ;

				return _xyz[i]; 
			}

		private:
			T _xyz[3];  //!< The 3 vector components

	}; // interface of T_Vector3


	// -------------------------------------------------------------------------- //
	//                                                                            //
	//  CLASS IMPLEMENTATION                                                      //
	//                                                                            //
	// -------------------------------------------------------------------------- //
	// -----------------------------------------------------------------------------
	//! Default constructor
	template<class T> T_Vector3<T>::T_Vector3 ()
	{
		_xyz[0] = 0.0;
		_xyz[1] = 0.0;
		_xyz[2] = 0.0;
	}

	// -----------------------------------------------------------------------------
	//! Default destructor
	template<class T> T_Vector3<T>::~T_Vector3 ()
	{
	}

	// -----------------------------------------------------------------------------
	//! Copy constructor
	template<class T> T_Vector3<T>::T_Vector3 (const T_Vector3& u)
	{
		setXYZ(u[0],u[1],u[2]);
	}

	// -----------------------------------------------------------------------------
	//! Copy constructor from a normalized vector
	template<class T> T_Vector3<T>::T_Vector3 (const T_NVector3<T>& u)
	{
		setXYZ(u[0],u[1],u[2]);
	}

	// -----------------------------------------------------------------------------
	//! Create a vector from real values
	template<class T> T_Vector3<T>::T_Vector3 (T x,T y,T z)
	{
		setXYZ(x,y,z);
	}
	// -----------------------------------------------------------------------------
	//! Assignment with a normalized vector
	template<class T> T_Vector3<T>& T_Vector3<T>::operator= (const T_NVector3<T>& u)
	{
		_xyz[0] = u[0];
		_xyz[1] = u[1];
		_xyz[2] = u[2];
		return ( *this );
	}

	// -----------------------------------------------------------------------------
	//! Self addition with a normalized vector
	template<class T> T_Vector3<T>& T_Vector3<T>::operator+= (const T_NVector3<T>& u)
	{
		_xyz[0] += u[0];
		_xyz[1] += u[1];
		_xyz[2] += u[2]; 
		return ( *this );
	}

	// -----------------------------------------------------------------------------
	//! Self substraction with a normalized vector
	template<class T> T_Vector3<T>& T_Vector3<T>::operator-= (const T_NVector3<T>& u)
	{
		_xyz[0] -= u[0];
		_xyz[1] -= u[1];
		_xyz[2] -= u[2]; 
		return ( *this );
	}


	// -----------------------------------------------------------------------------
	//! Inverse
	template<class T> T_Vector3<T> operator- (const T_Vector3<T>& u)
	{
		return ( T_Vector3<T>(-u._xyz[0], -u._xyz[1], -u._xyz[2]) );
	}


	// -----------------------------------------------------------------------------
	//! Left multiplication by a real value
	template<class T> T_Vector3<T> operator* (T r,const T_Vector3<T>& u)
	{
		return ( T_Vector3<T>(r*u._xyz[0], r*u._xyz[1], r*u._xyz[2]) );
	}


	// -----------------------------------------------------------------------------
	//! Norm
	template<class T> T T_Vector3<T>::norm () const
	{
		return sqrt( _xyz[0]*_xyz[0] + _xyz[1]*_xyz[1] + _xyz[2]*_xyz[2] );
	}

	// -----------------------------------------------------------------------------
	//! Square norm (self dot product)
	template<class T> T T_Vector3<T>::squareNorm () const
	{
		return _xyz[0]*_xyz[0] + _xyz[1]*_xyz[1] + _xyz[2]*_xyz[2];
	}

	// -----------------------------------------------------------------------------
	//! Infinite norm 
	template<class T> T T_Vector3<T>::infNorm() const
	{
		return std::max(std::max(fabs(_xyz[0]),fabs(_xyz[1])),fabs(_xyz[2])) ;
	}


	// -----------------------------------------------------------------------------
	//! Out stream override: prints the 3 vector components
	template<class T> std::ostream& operator<< (std::ostream& out,const T_Vector3<T>& u)
	{
		out << u[0] << " " << u[1] << " " << u[2];
		return ( out );
	}

	template<class T> T_Vector3<T> min(const T_Vector3<T>& v1,const T_Vector3<T>& v2) 
	{
		return T_Vector3<T>(std::min(v1[0],v2[0]),std::min(v1[1],v2[1]),std::min(v1[2],v2[2])) ;
	}

	template<class T> T_Vector3<T> max(const T_Vector3<T>& v1,const T_Vector3<T>& v2) 
	{
		return T_Vector3<T>(std::max(v1[0],v2[0]),std::max(v1[1],v2[1]),std::max(v1[2],v2[2])) ;
	}
}

#endif // _VECTOR3_H
